/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.test;

import java.util.UUID;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.network.VLANSupport;
import org.dasein.cloud.network.Vlan;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

public class VLANTestCase extends BaseTestCase {
    private CloudProvider cloud        = null;
    private String        testVlan     = null;
    private String        vlanToRemove = null;
    
    public VLANTestCase(String name) { super(name); }
    
    @Before
    public void setUp() throws InstantiationException, IllegalAccessException, CloudException, InternalException {
        String name = getName();
        
        cloud = getProvider();
        cloud.connect(getTestContext());
        if( name.equals("testGetVlan") || name.equals("testVlanContent") ) {
            for( Vlan vlan : cloud.getNetworkServices().getVlanSupport().listVlans() ) {
                testVlan = vlan.getProviderVlanId();
                break;
            }
            if( testVlan == null ) {
                vlanToRemove = cloud.getNetworkServices().getVlanSupport().provisionVlan("dsngettest-" + System.currentTimeMillis(), "DSN Get Test");
                testVlan = vlanToRemove;
            }
        }
        else if( name.equals("testRemoveVlan") ) {
            vlanToRemove = cloud.getNetworkServices().getVlanSupport().provisionVlan("dsnrmtest-" + System.currentTimeMillis(), "DSN Remove Test");
            testVlan = vlanToRemove;            
        }
    }
    
    @After
    public void tearDown() {
        try {
            if( vlanToRemove != null ) {
                cloud.getNetworkServices().getVlanSupport().removeVlan(vlanToRemove);
            }
        }
        catch( Throwable ignore ) {
            // ignore me
        }
        try {
            if( cloud != null ) {
                cloud.close();
            }
        }
        catch( Throwable ignore ) {
            // ignore
        }
    }
    
    @Test
    public void testGetBogusVlan() throws CloudException, InternalException {
        begin();
        Vlan vlan = cloud.getNetworkServices().getVlanSupport().getVlan(UUID.randomUUID().toString());
        
        assertNull("Found a matching VLAN for test", vlan);
        end();        
    }
    
    @Test
    public void testGetVlan() throws CloudException, InternalException {
        begin();
        Vlan vlan = cloud.getNetworkServices().getVlanSupport().getVlan(testVlan);
        
        assertNotNull("VLAN was not found", vlan);
        assertEquals("VLAN did not match target", testVlan, vlan.getProviderVlanId());
        out(vlan.toString());
        end();
    }
    
    @Test
    public void testListVlans() throws CloudException, InternalException {
        begin();
        VLANSupport vlanSupport = cloud.getNetworkServices().getVlanSupport();
        Iterable<Vlan> vlans = vlanSupport.listVlans();
        
        assertNotNull("VLAN list cannot be null", vlans);
        try {
            for( Vlan vlan : vlans ) {
                out("VLAN: " + vlan);
            }
        }
        catch( Throwable notPartOfTest ) {
            // ignore this
        }
        end();
    }
    
    @Test
    public void testMetaData() throws CloudException, InternalException {
        begin();
        VLANSupport support = cloud.getNetworkServices().getVlanSupport();
        
        out("Allows VLAN Creation:    " + support.allowsNewVlanCreation());
        out("Max VLAN Count:          " + support.getMaxVlanCount());
        out("Supports Custom Subnets: " + support.supportsCustomSubnets());
        end();
    }
    
    @Test
    public void testProvisionVlan() throws CloudException, InternalException {
        begin();
        if( cloud.getNetworkServices().getVlanSupport().allowsNewVlanCreation() ) {
            vlanToRemove = cloud.getNetworkServices().getVlanSupport().provisionVlan("dsntest-" + System.currentTimeMillis(), "Dasein Cloud Test VLAN");
            assertNotNull("Did not return any VLAN ID", vlanToRemove);
            out(vlanToRemove);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            Vlan vlan = cloud.getNetworkServices().getVlanSupport().getVlan(vlanToRemove);
            
            assertNotNull("Not able to find newly created vlan", vlan);
            assertEquals("Created vlan ID does not match", vlanToRemove, vlan.getProviderVlanId());
        }
        else {
            try {
                vlanToRemove = cloud.getNetworkServices().getVlanSupport().provisionVlan("dsntest-" + System.currentTimeMillis(), "Dasein Cloud Test VLAN");
                fail("Implementations that do not support VLAN creation should throw OperationNotSupportedException");
            }
            catch( OperationNotSupportedException e ) {
                // expected
            }
        }
        end();
    }
    
    @Test
    public void testProvisionVlanWithSubnet() throws CloudException, InternalException {
        begin();
        if( cloud.getNetworkServices().getVlanSupport().supportsCustomSubnets() ) {
            vlanToRemove = cloud.getNetworkServices().getVlanSupport().provisionVlanWithSubnet("dsntest-" + System.currentTimeMillis(), "Dasein Cloud Test VLAN", "10.99.99.12/24", "10.99.99.254", "10.99.99.253", "10.99.99.252");
            assertNotNull("Did not return any VLAN ID", vlanToRemove);
            out(vlanToRemove);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            Vlan vlan = cloud.getNetworkServices().getVlanSupport().getVlan(vlanToRemove);
            
            assertNotNull("Not able to find newly created vlan", vlan);
            assertEquals("Created vlan ID does not match", vlanToRemove, vlan.getProviderVlanId()); 
        }
        else {
            try {
                vlanToRemove = cloud.getNetworkServices().getVlanSupport().provisionVlanWithSubnet("dsntest-" + System.currentTimeMillis(), "Dasein Cloud Test VLAN", "10.99.99.12/24", "10.99.99.254", "10.99.99.253", "10.99.99.252");
                fail("Implementations that do not support custom subnets should throw an OperationNotSupportedException");
            }
            catch( OperationNotSupportedException e ) {
                // expected
            }
        }
        end();
    }
    
    @Test
    public void testRemoveVlan() throws CloudException, InternalException {
        begin();
        cloud.getNetworkServices().getVlanSupport().removeVlan(vlanToRemove);
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        Vlan vlan = cloud.getNetworkServices().getVlanSupport().getVlan(vlanToRemove);

        assertNull("VLAN was not removed", vlan);
        vlanToRemove = null;
        end();
    }
    
    @Test
    public void testVlanContent() throws CloudException, InternalException {
        begin();
        Vlan vlan = cloud.getNetworkServices().getVlanSupport().getVlan(testVlan);
        
        assertNotNull("VLAN was not found", vlan);
        assertEquals("VLAN did not match target", testVlan, vlan.getProviderVlanId());
        assertNotNull("VLAN has no name", vlan.getName());
        assertNotNull("VLAN has no description", vlan.getDescription());
        assertEquals("VLAN region mismatch", cloud.getContext().getRegionId(), vlan.getProviderRegionId());
        assertNotNull("VLAN has not data center", vlan.getProviderDataCenterId());
        out("ID:          " + vlan.getProviderVlanId());
        out("Name:        " + vlan.getName());
        out("Owner:       " + vlan.getProviderOwnerId());
        out("Region:      " + vlan.getProviderRegionId());
        out("Data Center: " + vlan.getProviderDataCenterId());
        out("CIDR:        " + vlan.getCidr());
        out("Gateway:     " + vlan.getGateway());
        end();        
    }
}
